<?php

namespace Drp\DynamicAttributes\Console\Commands;

use Drp\DynamicAttributes\AttributeSchema;
use Drp\DynamicAttributes\DynamicModel;
use Illuminate\Console\Command;

class DynamicSchemaCommand extends Command
{
    /**
     * The name and signature of the console command.
     *
     * @var string
     */
    protected $signature = 'dynamic:schema {class : Model class to check}';

    /**
     * The console command description.
     *
     * @var string
     */
    protected $description = 'Command line manager for dynamic attributes';

    /**
     * Class to look at schema for.
     *
     * @var string
     */
    protected $class;

    /**
     * Execute the console command.
     *
     * @return mixed
     */
    public function handle()
    {
        $this->class = trim($this->argument('class'), '\\');
        if (!class_exists($this->class)) {
            $this->error('Class ' . $this->class . ' does not exist');

            return;
        }

        try {
            $model = app($this->class);
            if (!($model instanceof DynamicModel)) {
                $this->error('Class given is not a DynamicAttributes model');
            }
        } catch (\Exception $exception) {
            $this->error(
                'Unable to verify if ' .
                $this->class .
                ' is a DynamicAttributes Model'
            );

            $continue = $this->confirm(
                'Would you like to continue anyway?',
                false
            );

            if (!$continue) {
                return;
            }
        }

        $this->start();
    }

    /**
     * Function to start the schema manipulation
     *
     * @return void
     */
    public function start()
    {
        $choice = $this->choice(
            'What would you like to do?',
            [
                'Add schema',
                'View schema',
                'Edit schema',
                'Remove schema',
                'Exit',
            ],
            '1'
        );

        if ($choice === 'Exit') {
            return;
        }

        $function = camel_case($choice);

        $this->{$function}();
    }

    /**
     * Asking the user the attributes for the new schema
     *
     * @return void
     */
    public function addSchema()
    {
        $schema = new AttributeSchema();
        $schema = $this->updateSchemaColumns($schema);
        if ($schema === false) {
            return;
        }

        $this->info('Successfully created ' . $schema->field . ' schema');
        $this->start();
    }

    /**
     * Showing the user the schema that currently exists on the model
     *
     * @return void
     */
    public function viewSchema()
    {
        $modelSchema = $this->getSchema();
        if ($modelSchema === false) {
            $this->start();
        }

        $fields = [
            'field',
            'type',
            'allow_null',
        ];

        $schemaData = [];

        foreach ($modelSchema as $schema) {
            $reduced      = array_intersect_key($schema->toArray(), array_flip($fields));
            $schemaData[] = $reduced;
        }

        $this->table($fields, $schemaData);

        $this->start();
    }

    /**
     * Asking the user what schema they would like to edit
     *
     * @return void
     */
    public function editSchema()
    {
        $result = $this->getSchemaChoices();
        if ($result === false) {
            $this->start();

            return;
        }

        $this->editChoices(
            $result['choices'],
            $result['dangerousChoices'],
            $result['modelSchema']
        );
    }

    /**
     * Asking the user which schema they would like to remove
     *
     * @return void
     */
    public function removeSchema()
    {
        $result = $this->getSchemaChoices();

        if ($result === false) {
            $this->start();

            return;
        }

        $schemaChoice = $this->choice(
            'Which schema would you like to delete?',
            $result['choices']
        );

        if ($schemaChoice == 'None') {
            $this->start();

            return;
        }

        $dangerousChoice = in_array($schemaChoice, $result['dangerousChoices']);

        if ($dangerousChoice) {
            $confirmDangerous = $this->confirm(
                'You are going to remove a schema that already has values ' .
                'associated, removing this will also remove those values',
                false
            );

            if (!$confirmDangerous) {
                $this->removeSchema();
            }
        }

        $confirmDelete = $this->confirm(
            'Are you sure you want to delete ' . $schemaChoice . '?',
            false
        );

        if ($confirmDelete === false) {
            $this->removeSchema();

            return;
        }

        $schema = $result['modelSchema']->keyBy('field')->get($schemaChoice);
        $schema->delete();

        $this->start();
    }

    /**
     * Get the schema for the class
     *
     * @return AttributeSchema
     */
    protected function getSchema()
    {
        $modelSchema = AttributeSchema::where('model', $this->class)->get();
        if ($modelSchema->count() === 0) {
            $this->warn('No schema found for ' . $this->class);

            return false;
        }

        return $modelSchema;
    }

    /**
     * Gets the schema choices in a formatted result and filter out schema
     * that shouldn't be changed
     *
     * @return array
     */
    protected function getSchemaChoices()
    {
        $modelSchema     = $this->getSchema();

        if ($modelSchema == null) {
            return false;
        }

        $schemaChoices   = [];
        $filteredChoices = [];
        foreach ($modelSchema as $schema) {
            if ($schema->attributes->count() > 0) {
                $filteredChoices[] = $schema->field;
            } else {
                $schemaChoices[] = $schema->field;
            }
        }

        if (count($filteredChoices) > 0) {
            $this->warn(
                implode(', ', $filteredChoices) .
                ' are dangerous to change as they have values associated to them'
            );
        }

        $schemaChoices   = array_merge($schemaChoices, $filteredChoices);
        $schemaChoices[] = 'None';

        return [
            'choices'          => $schemaChoices,
            'dangerousChoices' => $filteredChoices,
            'modelSchema'      => $modelSchema,
        ];
    }

    /**
     * Updates the required columns on a schema
     *
     * @param  AttributeSchema $schema
     * @return AttributeSchema
     */
    protected function updateSchemaColumns($schema)
    {
        if ($schema === null) {
            $this->error('Can not update a empty schema');

            return;
        }

        $field = $this->getFieldName($schema, $schema->field);

        if ($field === false) {
            $this->start();

            return false;
        }

        $this->info('Field name has been set to ' . $field);

        $default = $this->getDefaultValue('type', $schema);
        $types   = [
            gettype(''),
            gettype(0),
            gettype(0.0),
            gettype(false),
        ];
        $default = array_search($default, $types);
        $type    = $this->choice('What is the type?', $types, $default);
        $this->info('Type has been set to ' . $type);

        $default   = $this->getDefaultValue('allow_null', $schema, true);
        $allowNull = $this->confirm('Can it be null?', $default);

        $schema->fill([
            'field'      => $field,
            'type'       => $type,
            'allow_null' => $allowNull,
            'model'      => $this->class,
        ]);

        $schema->save();

        return $schema;
    }

    /**
     * Gets user to give a field name, if it already exists gives user a choice
     * to cancel or try again.
     *
     * @param  AttributeSchema $schema
     * @param  string          $current
     * @return mixed
     */
    protected function getFieldName($schema, $current = '')
    {
        $default      = $this->getDefaultValue('field', $schema);
        $field        = $this->ask('What is the field name?', $default);
        $field        = str_replace(' ', '_', strtolower($field));
        $allSchema    = $this->getSchema();
        $schemaFields = [];

        if ($allSchema != null) {
            $schemaFields = $allSchema->pluck('field')->toArray();
        }

        if (in_array($field, $schemaFields) && $field != $current) {
            $this->error('Schema with that field name already exists');
            $retry = $this->confirm('Would you like to retry?');
            if ($retry) {
                return $this->getFieldName($schema, $current);
            }

            return false;
        }

        return $field;
    }

    /**
     * Gets the value from the item, if it doesn't exist return the default
     *
     * @param string $columnName Name of column to get
     * @param string $item       Item to get value from
     * @param mixed  $default    Default to result if column doesn't exist
     *
     * @return mixed
     */
    public function getDefaultValue($columnName, $item, $default = null)
    {
        $value = $item->{$columnName};

        if ($value != '') {
            return $value;
        }

        return $default;
    }

    /**
     * Ask the user which schema they would like to edit
     *
     * @param array           $schemaChoices
     * @param array           $dangerousChoices
     * @param AttributeSchema $modelSchema
     *
     * @return void
     */
    protected function editChoices(
        $schemaChoices,
        $dangerousChoices,
        $modelSchema
    ) {
        $schemaChoice = $this->choice(
            'Which schema would you like to edit?',
            $schemaChoices
        );

        if ($schemaChoice == 'None') {
            $this->start();

            return;
        }

        $dangerousChoice = in_array($schemaChoice, $dangerousChoices);

        if ($dangerousChoice) {
            $confirmDangerous = $this->confirm(
                'You are going to edit a schema that already has values ' .
                'associated, editing this might affect the values associated',
                false
            );

            if (!$confirmDangerous) {
                $this->editChoices(
                    $schemaChoices,
                    $dangerousChoices,
                    $modelSchema
                );
            }
        }

        $schema = $modelSchema->keyBy('field')->get($schemaChoice);
        $result = $this->updateSchemaColumns($schema);
        if ($result === false) {
            return;
        }

        $this->info($schema->field . ' has been successfully updated');

        $result = $this->getSchemaChoices();
        $this->editChoices(
            $result['choices'],
            $result['dangerousChoices'],
            $result['modelSchema']
        );
    }
}
